//
// Scheduler.cpp
//
// Copyright (c) Shareaza Development Team, 2002-2007.
// This file is part of SHAREAZA (shareaza.sourceforge.net)
//
// Shareaza is free software; you can redistribute it
// and/or modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// Shareaza is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Shareaza; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//

#include "StdAfx.h"
#include "atltime.h"
#include "Shareaza.h"
#include "Settings.h"
#include "DlgHelp.h"
#include "Scheduler.h"
#include "Network.h"
#include "Security.h"
#include "Ras.h"
#include "Reason.h"

CScheduler Scheduler;

//////////////////////////////////////////////////////////////////////
// CScheduler load and save

bool CScheduler::Load()
{
	CQuickLock oLock( Scheduler.m_pSection );
	CFile pFile;
	CString strFile = Settings.General.UserPath + _T("\\Data\\SchTasks.dat");

	// Try to open the scheduler file
	if ( ! pFile.Open( strFile, CFile::modeRead ) ) 
	{
		theApp.Message( MSG_ERROR, _T("Failed to open SchTasks.dat") );
		return FALSE;
	}

	try
	{
		CArchive ar( &pFile, CArchive::load );	// 4 KB buffer

		Serialize( ar );
		ar.Close();
	}
	catch ( CException* pException )
	{
		pException->Delete();
	}

	pFile.Close();

	return TRUE;
}

bool CScheduler::Save()
{
	CQuickLock oLock( Scheduler.m_pSection );
	CString strFile = Settings.General.UserPath + _T("\\Data\\SchTasks.dat");

	CFile pFile;
	if ( ! pFile.Open( strFile, CFile::modeWrite|CFile::modeCreate ) )
		return FALSE;

	try
	{
		CArchive ar( &pFile, CArchive::store );	// 4 KB buffer
		try
		{
			Serialize( ar );
			ar.Close();
		}
		catch ( CException* pException )
		{
			ar.Abort();
			pFile.Abort();
			pException->Delete();
			return FALSE;
		}
		pFile.Close();
	}
	catch ( CException* pException )
	{
		pFile.Abort();
		pException->Delete();
		return FALSE;
	}


	return TRUE;
}

//////////////////////////////////////////////////////////////////////
// CScheduler serialize

void CScheduler::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// Read the number of scheduled tasks
		ar << ScheduleList.GetCount();
		for ( POSITION pos = GetIterator() ; pos ; )
		{
			// Get a pointer to each task
			CScheduleItem *pItem = GetNext(pos);
			// Store each task's data
			pItem->Serialize(ar,1);
		}
	}
	else
	{
		// First clear any existing tasks
		Clear();
		int nNumTasks;
		// Read the number of tasks to load
		for (ar >> nNumTasks; nNumTasks > 0; nNumTasks--)
		{
			// Create a new instance of each task
			CScheduleItem *pItem = new CScheduleItem();
			// Read each task's data
			pItem->Serialize(ar,1);
			// Add the task to the task list
			ScheduleList.AddTail(pItem);
		}
	}
}

//////////////////////////////////////////////////////////////////////
// CScheduleItem construction

CScheduleItem::CScheduleItem()
{
	m_bEveryDay = false;
	m_nAction = 0;
	m_sDescription = "";
	m_tScheduleDateTime = 0;

	m_bActive = false;
	m_bExecuted = false;
}

CScheduleItem::~CScheduleItem()
{
}

//////////////////////////////////////////////////////////////////////
// CScheduleItem class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// CScheduleItem serialize

void CScheduleItem::Serialize(CArchive& ar, int /*nVersion*/)
{
	if ( ar.IsStoring() )
	{
		// Store all task variables
		ar << m_bEveryDay;
		ar << m_nAction;
		ar << m_sDescription;
		ar << m_tScheduleDateTime;
		ar << m_sDate;
		ar << m_sTime;
		ar << m_bActive;
		ar << m_bExecuted;
		ar << m_nLimited;
		ar << m_nLimitedDown;
		ar << m_nLimitedUp;
		ar << m_bToggleBandwidth;
		ar << m_bLimitedNetworks;
	}
	else
	{
		// Load all task variables
		ar >> m_bEveryDay;
		ar >> m_nAction;
		ar >> m_sDescription;
		ar >> m_tScheduleDateTime;
		ar >> m_sDate;
		ar >> m_sTime;
		ar >> m_bActive;
		ar >> m_bExecuted;
		ar >> m_nLimited;
		ar >> m_nLimitedDown;
		ar >> m_nLimitedUp;
		ar >> m_bToggleBandwidth;
		ar >> m_bLimitedNetworks;
	}
}


//////////////////////////////////////////////////////////////////////
// CScheduler class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// CScheduler construction

CScheduler::CScheduler()
{
}

CScheduler::~CScheduler()
{
	Clear();
}

//////////////////////////////////////////////////////////////////////
// CScheduler task list access

POSITION CScheduler::GetIterator() const
{
	return ScheduleList.GetHeadPosition();
}

CScheduleItem* CScheduler::GetNext(POSITION& pos) const
{
	return (CScheduleItem*)ScheduleList.GetNext( pos );
}

int CScheduler::GetCount()
{
	return ScheduleList.GetCount();
}

//Checks to see if pRule exists in the main list
BOOL CScheduler::Check(CScheduleItem* pRule) const
{
	CQuickLock oLock( Scheduler.m_pSection );	//This means no task will be executed while dialog is open
	return pRule && ( ScheduleList.Find( pRule ) != NULL );
}


//////////////////////////////////////////////////////////////////////
// CScheduler item modification

void CScheduler::Add(CScheduleItem* pScheduleItem)
{
	CQuickLock oLock(m_pSection);
	POSITION pos = ScheduleList.Find( pScheduleItem );
	if ( pos == NULL ) ScheduleList.AddHead( pScheduleItem );
}

void CScheduler::Remove(CScheduleItem* pRule)
{
	CQuickLock oLock(m_pSection);
	POSITION pos = ScheduleList.Find( pRule );
	if ( pos ) ScheduleList.RemoveAt( pos );
	delete pRule;
}

void CScheduler::Clear()
{
	CQuickLock oLock(m_pSection);
	for ( POSITION pos = GetIterator() ; pos ; )
	{
		delete GetNext( pos );
	}

	ScheduleList.RemoveAll();
}


//////////////////////////////////////////////////////////////////////
// CScheduler schedule checking

void CScheduler::CheckSchedule()
{
	static DWORD LastTicks = 0;
	DWORD Ticks = GetTickCount();
	// Only check after a second has passed
	if ((Ticks - LastTicks) > 1000) 
	{
		CTime tNow = CTime::GetCurrentTime();
		CQuickLock oLock(m_pSection);
		for ( POSITION pos = GetIterator() ; pos ; )
		{
			CScheduleItem *pItem = GetNext(pos);
			// Check if a task should be executed
			if (pItem->m_bActive)
			{
				if(ScheduledTimePassed(&tNow,&pItem->m_tScheduleDateTime))
				{
					// Execute the task
					theApp.m_bBusy = true;
					
					DoScheduledItem(pItem);
					// If everyday, schedule task for tomorrow
					if (pItem->m_bEveryDay)
						pItem->m_tScheduleDateTime += CTimeSpan::CTimeSpan(1,0,0,0);
					else
						pItem->m_bActive = false;

					pItem->m_bExecuted = true;
					Scheduler.Save();
					theApp.m_bBusy = false;
					
				}
				else
				{
					Settings.Scheduler.Enable = true;	//replacement for IsEnabled()
				}
			}
		}
		
		LastTicks = Ticks;
	}
}

//////////////////////////////////////////////////////////////////////
// CScheduler execute task

void CScheduler::DoScheduledItem(CScheduleItem *pRule)
{
	// Execute the selected scheduled task
	switch ( pRule->m_nAction )
	{
	case SCHEDULE_FULL_SPEED:  // Set the bandwidth to full speed
		theApp.Message( MSG_DEBUG, _T("Scheduler - Bandwidth: Full Speed") );
		Settings.Live.BandwidthScale = 100;
		Settings.Gnutella2.EnableToday	= TRUE;
		Settings.Gnutella1.EnableToday	= Settings.Gnutella1.EnableAlways;
		Settings.eDonkey.EnableToday	= Settings.eDonkey.EnableAlways;
		if ( ! Network.IsConnected() ) Network.Connect( TRUE );
		break;
	case SCHEDULE_LIMITED_SPEED:  // Set the bandwidth to limited speed
		theApp.Message( MSG_DEBUG, _T("Scheduler - Bandwidth: Limited Speed") );

		Settings.Scheduler.LimitedNetworks	= pRule->m_bLimitedNetworks;
		if (!pRule->m_bToggleBandwidth)
		{
			Settings.Live.BandwidthScale = pRule->m_nLimited;
		}
		else
		{
			Settings.Bandwidth.Downloads = ( (Settings.Bandwidth.Downloads * pRule->m_nLimitedDown)/100 );
			Settings.Bandwidth.Uploads = ( (Settings.Bandwidth.Uploads * pRule->m_nLimitedUp)/100 );
		}
		Settings.Gnutella2.EnableToday	= TRUE;
		Settings.Gnutella1.EnableToday	= pRule->m_bLimitedNetworks ? FALSE :Settings.Gnutella1.EnableAlways;
		Settings.eDonkey.EnableToday	= pRule->m_bLimitedNetworks ? FALSE :Settings.eDonkey.EnableAlways;
		if ( ! Network.IsConnected() ) Network.Connect( TRUE );
		break;
	case SCHEDULE_OFF:	// Set the bandwidth to 0 and disconnect all networks
		theApp.Message( MSG_DEBUG, _T("Scheduler - Bandwidth: Stop") );
		Settings.Live.BandwidthScale = 0;
		Settings.Gnutella2.EnableToday	= FALSE;
		Settings.Gnutella1.EnableToday	= FALSE;
		Settings.eDonkey.EnableToday	= FALSE;
		if ( Network.IsConnected() ) Network.Disconnect();
		break;
	case SCHEDULE_DISCONNECT:	// Disconnect
		theApp.Message( MSG_DEBUG, _T("Scheduler - System: Disconnect Dial-up Connection") );
		Settings.Live.BandwidthScale = 0;
		Settings.Gnutella2.EnableToday	= FALSE;
		Settings.Gnutella1.EnableToday	= FALSE;
		Settings.eDonkey.EnableToday	= FALSE;
		if ( Network.IsConnected() ) Network.Disconnect();
		HangUpConnection(); // Hang up the connection
		break;
	case SCHEDULE_EXIT:		// Exit Shareaza
		theApp.Message( MSG_DEBUG, _T("Scheduler - System: Exit Shareaza") );
		if(!PostMainWndMessage( WM_CLOSE)) theApp.Message(MSG_ERROR, _T("Scheduler failed to send CLOSE message"));

		break;
	case SCHEDULE_SHUTDOWN:	// Shut down the computer
		theApp.Message( MSG_DEBUG, _T("Scheduler - System: Shut Down the Computer") );

		// If we dont have shutdown rights
		if (!Settings.Scheduler.ShutdownRights) 
		{
			theApp.Message( MSG_DEBUG, _T("Insufficient rights to shut down the system") );
			return;
		}
		if (ShutDownComputer())// Close Shareaza if shutdown successfully started
			if(!PostMainWndMessage( WM_CLOSE)) theApp.Message(MSG_ERROR, _T("Scheduler failed to send CLOSE message"));

		else theApp.Message( MSG_DEBUG, _T("System shutdown failed!"));

		break;
	default: //Error
		theApp.Message( MSG_ERROR, _T("Invalid value in schedule") );
		break;
	}
}

// Check if any tasks are scheduled
//bool CScheduler::IsEnabled()
//{
//	POSITION pos = GetIterator();
//	bool bActiveFound = false;
//	while (pos && !bActiveFound)
//	{
//		CScheduleItem *pItem = GetNext( pos );
//		if (pItem->m_bActive)
//		{
//			bActiveFound = true;
//		}
//	}
//
//	return bActiveFound;
//}

// Disconnect a dial-up connection
void CScheduler::HangUpConnection()
{
	DWORD dwCb = sizeof(RASCONN);
	DWORD dwConnections = 0;
	RASCONN* lpRasConn = NULL;
	LPRASCONNSTATUS RasConStatus = 0;

	// Allocate the size needed for the RAS structure.
	lpRasConn = (RASCONN*)HeapAlloc(GetProcessHeap(), 0, dwCb);

	// Set the structure size for version checking purposes.
	lpRasConn->dwSize = sizeof(RASCONN);

	// Call the RAS API 
	RasEnumConnections(lpRasConn,&dwCb, &dwConnections);

	DWORD i;
	int loop = 0;

	for (i = 0; i < dwConnections; i++) // Loop through all current connections
	{
		RasHangUp(lpRasConn[i].hrasconn); // Hang up the connection
		while((RasGetConnectStatus(lpRasConn[i].hrasconn,RasConStatus) || (loop > 30)))
		{
			// Loop until the connection handle is invalid, or 3 seconds have passed
			Sleep(100);
			loop++;
		}
	}

	// Free the memory if necessary.
	if (lpRasConn != NULL)
	{
		HeapFree(GetProcessHeap(), 0, lpRasConn);
		lpRasConn = NULL;
	}
}

bool CScheduler::ShutDownComputer()
{
	int ShutdownSuccess = 0;
	UINT ShutdownFlags = (EWX_POWEROFF | EWX_FORCEIFHUNG); // Force only hung programs to close (not so safe)
	//UINT ShutdownFlags = (EWX_POWEROFF | EWX_FORCE); // Force all programs to close (safer)

	// Try 2000/XP way first
	ShutdownSuccess = InitiateSystemShutdownEx(NULL,_T("Shareaza Scheduled Shutdown\n\nA system shutdown was scheduled using Shareaza. The system will now shut down."),30,TRUE,FALSE,SHTDN_REASON_FLAG_USER_DEFINED);
	// Fall back to 9x way if this does not work
	if (!ShutdownSuccess)
	{
		DWORD dReason;
		dReason = (SHTDN_REASON_MAJOR_OTHER | SHTDN_REASON_MINOR_OTHER | SHTDN_REASON_FLAG_PLANNED);
		ShutdownSuccess = ExitWindowsEx(ShutdownFlags,dReason);
	}
	return (ShutdownSuccess != 0);
}

// Give the process shutdown rights
bool CScheduler::SetShutdownRights()
{
	HANDLE hToken; 
	TOKEN_PRIVILEGES tkp; 

	// Get a token for this process. 
	if (!OpenProcessToken(GetCurrentProcess(), 
		TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) 
		return( FALSE ); 

	// Get the LUID for the shutdown privilege. 
	LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, 
		&tkp.Privileges[0].Luid); 

	tkp.PrivilegeCount = 1;  // one privilege to set    
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 

	// Get the shutdown privilege for this process. 
	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, 
		(PTOKEN_PRIVILEGES)NULL, 0); 

	if (GetLastError() != ERROR_SUCCESS) 
		return FALSE; 

	return TRUE;
}

bool CScheduler::ScheduledTimePassed(CTime *pNow, CTime *pThen)
{
	/*
	Mathematical way, faster?

	long SecsNow, SecsScheduled;
	SecsNow = (pNow->wYear - 2001) * 365 * 24 * 60 * 60;
	SecsNow += (pNow->wDay - 1) * 24 * 60 * 60;
	SecsNow += (pNow->wHour - 1) * 60 * 60;
	SecsNow += (pNow->wMinute - 1) * 60;
	SecsNow += (pNow->wSecond);

	SecsScheduled = (pThen->wYear - 2001) * 365 * 24 * 60 * 60;
	SecsScheduled += (pThen->wDay - 1) * 24 * 60 * 60;
	SecsScheduled += (pThen->wHour - 1) * 60 * 60;
	SecsScheduled += (pThen->wMinute - 1) * 60;
	SecsScheduled += (pThen->wSecond);

	*/

	// Logical way, fewer calculations?

	if (pNow->GetYear() > pThen->GetYear())
		return true;
	else
		if (pNow->GetYear() == pThen->GetYear())
			if (pNow->GetMonth() > pThen->GetMonth())
				return true;
			else
				if (pNow->GetMonth() == pThen->GetMonth())
					if (pNow->GetDay() > pThen->GetDay())
						return true;
					else
						if (pNow->GetDay() == pThen->GetDay())
							if (pNow->GetHour() > pThen->GetHour() )
								return true;
							else
								if (pNow->GetHour()  == pThen->GetHour() )
									if (pNow->GetMinute() > pThen->GetMinute())
										return true;
									else
										if (pNow->GetMinute() == pThen->GetMinute())
											if (pNow->GetSecond() > pThen->GetSecond())
												return true;
	return false;


	//return SecsNow > SecsScheduled;
}

//Calculates the different between current hour and shutdown hour today
//Caller must first check to see if scheduler is enable or not
int CScheduler::GetHoursToShutdown()
{
	int nHoursToShutdown = 0xFF;		//Means no Shutdown or disconnect is scheduled for today
	POSITION pos = GetIterator();
	CQuickLock oLock(m_pSection);
	while (pos)
	{
		CScheduleItem *pItem = GetNext( pos );
		if ( pItem->m_bActive )
		{
			if(pItem->m_nAction == SCHEDULE_OFF || 
				pItem->m_nAction == SCHEDULE_DISCONNECT || 
				pItem->m_nAction == SCHEDULE_EXIT || 
				pItem->m_nAction == SCHEDULE_SHUTDOWN)
			{

				CTime tNow = CTime::GetCurrentTime();
				if( pItem->m_tScheduleDateTime.GetYear() == tNow.GetYear() &&
					pItem->m_tScheduleDateTime.GetMonth() == tNow.GetMonth() &&
					pItem->m_tScheduleDateTime.GetDay() == tNow.GetDay())
				{
					int nThisHoursToShutDown = pItem->m_tScheduleDateTime.GetHour() - tNow.GetHour();
					if( nThisHoursToShutDown < nHoursToShutdown && nThisHoursToShutDown > 0 ) nHoursToShutdown = nThisHoursToShutDown;
				}
			}
		}
	}

	return nHoursToShutdown;
}